#include "AbstractFlat.hpp"

#include <imgui.h> // NOLINT
#include <toml.h>
#include "../world/Chunk.hpp"
#include "../world/Area.hpp"

namespace contouring {
    size_t AbstractFlat::clear(const voxel_pos& pos, const world::area_map& areas) {
        size_t buffer_count = 0;
        auto it_a = buffers.begin();
        while (it_a != buffers.end()) { // Remove out of range buffers
            if (const auto area = areas.find(it_a->first); area != areas.end()) {
                //Update
                it_a->second.first = area->second->getOffset();
                const auto center = glm::divide(pos - it_a->second.first.as_voxel());
                auto &bfs = it_a->second.second;
                auto it = bfs.begin();
                while(it != bfs.end()) {
                    if (glm::length2(center - it->first) > glm::pow2(keepDistance)) {
                        if(it->second != NULL)
                            delete it->second;

                        it = bfs.erase(it);
                    } else {
                        buffer_count++;
                        ++it;
                    }
                }
                ++it_a;
            } else {
                for(auto& buffer: it_a->second.second) {
                    if(buffer.second != NULL)
                        delete buffer.second;
                }
                it_a = buffers.erase(it_a);
            }
        }
        return buffer_count;
    }

    AbstractFlat::AbstractFlat(const std::string& str): Abstract() {
        auto opt = toml::parse(str);
        loadDistance = opt["load_distance"].value_or(loadDistance);
        keepDistance = opt["keep_distance"].value_or(keepDistance);
    }
    std::string AbstractFlat::getOptions() const {
        std::ostringstream ss;
        ss << toml::table({
            {"load_distance", loadDistance},
            {"keep_distance", keepDistance}
        });
        return ss.str();
    }
    std::pair<float, float> AbstractFlat::getFarRange() const {
        return std::make_pair((loadDistance - 1.5) * CHUNK_LENGTH, (keepDistance + .5) * CHUNK_LENGTH);
    }

    void AbstractFlat::onGui() {
        ImGui::SliderInt("Load Distance", &loadDistance, 1, keepDistance);
        ImGui::SliderInt("Keep Distance", &keepDistance, loadDistance+1, 21);
    }

    void AbstractFlat::getModels(std::vector<std::pair<glm::mat4, buffer::Abstract *const>> &out, const std::optional<geometry::Frustum> &frustum, const glm::llvec3& offset, int density) {
        const auto scaling = glm::scale(glm::mat4(1), glm::vec3(1.f / density));
        for (const auto [_, area] : buffers) {
        for (const auto [pos, buffer] : area.second) {
            const glm::vec3 fPos = (glm::vec3(area.first.raw_as_long() + glm::multiply(pos) - offset * glm::llvec3(density)) + area.first.offset) / glm::vec3(density);
            if (buffer != NULL && (!frustum.has_value() || frustum.value().contains(geometry::Box::fromMin(fPos, glm::vec3(CHUNK_LENGTH / (float)density)))))
                out.emplace_back(glm::translate(scaling, fPos * (float)density), buffer);
        }}
    }
}