#include "FlatDualMC.hpp"

#include "../world/Chunk.hpp"
#include "../world/materials.hpp"
#include <Remotery.h>  // NOLINT
#include <imgui.h>     // NOLINT
#include <toml.h>
#include "dualmc.h"

namespace contouring {
    FlatDualMC::FlatDualMC(const std::string &str) : AbstractFlat(str) {
        auto opt = toml::parse(str);
        iso = opt["iso"].value_or(iso);
        manifold = opt["manifold"].value_or(manifold);

        for (size_t i = 1; i <= 2; i++) {
            workers.emplace_back([&] {
                while (running) {
                    std::pair<area_<chunk_pos>, surrounding::corners> ctx;
                    loadQueue.wait();
                    if (loadQueue.pop(ctx)) {
                        rmt_ScopedCPUSample(ProcessContouring, 0);
                        buffer::ShortIndexed::Data data;
                        render(ctx.second, data);
                        loadedQueue.emplace(ctx.first, data);
                    }
                }
            });
        }
    }
    FlatDualMC::~FlatDualMC() {
        running = false;
        loadQueue.notify_all();

        for(auto& worker: workers) {
            if (worker.joinable())
                worker.join();
        }
    }

    std::string FlatDualMC::getOptions() const {
        std::ostringstream ss;
        ss << toml::table({
            {"load_distance", loadDistance},
            {"keep_distance", keepDistance},
            {"iso", iso},
            {"manifold", manifold}
        });
        return ss.str();
    }

     void FlatDualMC::enqueue(const area_<chunk_pos> &pos, const chunk_pos &offset, const world::ChunkContainer &data) {
        rmt_ScopedCPUSample(EnqueueContouring, RMTSF_Aggregate);
        const auto dist2 = glm::length2(offset - pos.second);
        if (dist2 <= loadDistance * loadDistance) {
            surrounding::corners surrounding;
            if(surrounding::load(surrounding, pos.second, data)) {
                loadQueue.push(pos, surrounding, -dist2);
            }
        }
     }

    void FlatDualMC::onUpdate(const area_<chunk_pos> &pos, const chunk_pos &offset, const world::ChunkContainer &data, geometry::Faces neighbors) {
        enqueue(pos, offset, data);
        if (neighbors && (geometry::Faces::Left | geometry::Faces::Down | geometry::Faces::Backward)) {
            for (size_t i = 1; i < 8; i++) {
                enqueue(std::make_pair(pos.first, pos.second - surrounding::g_corner_offsets[i]), offset, data);
            }
        }
    }

    void FlatDualMC::onNotify(const area_<chunk_pos> &pos, const chunk_pos &offset, const world::ChunkContainer &data) {
        const auto it = buffers.find(pos.first);
        if(it == buffers.end() || it->second.second.find(pos.second) == it->second.second.end()) {
            enqueue(pos, offset, data);
        }
    }

    void FlatDualMC::update(const voxel_pos& pos, const world::area_map& areas) {
        std::pair<area_<chunk_pos>, buffer::ShortIndexed::Data> out;
        reports.load.push(loadQueue.size());
        //MAYBE: clear out of range loadQueue.trim(keepDistance * keepDistance)
        reports.loaded.push(loadedQueue.size());
        while(loadedQueue.pop(out)) {
            const auto buffer = new buffer::ShortIndexed(GL_TRIANGLES, out.second);
            auto &bfs = buffers[out.first.first].second; //NOTE: buffer.first uninitialized (will be set in clear())
            if (const auto it = bfs.find(out.first.second); it != bfs.end()) {
                if(it->second != NULL)
                    delete it->second;

                it->second = buffer;
            } else {
                bfs.emplace(out.first.second, buffer);
            }
        }
        size_t count = AbstractFlat::clear(pos, areas);
        reports.count.push(count);
    }

    void FlatDualMC::onGui() {
        ImGui::PlotHistogram("Count", reports.count.buffer.get(), reports.count.size, 0, std::to_string(reports.count.current()).c_str(), 0);
        ImGui::PlotHistogram("Loading", reports.load.buffer.get(), reports.load.size, 0, std::to_string(reports.load.current()).c_str(), 0);
        ImGui::PlotHistogram("Waiting", reports.loaded.buffer.get(), reports.loaded.size, 0, std::to_string(reports.loaded.current()).c_str(), 0);
        ImGui::Separator();
        AbstractFlat::onGui();
        ImGui::Separator();
        ImGui::SliderFloat("Iso", &iso, 0, 1);
        ImGui::Checkbox("Manifold", &manifold);
    }

    void FlatDualMC::render(const surrounding::corners &surrounding, buffer::ShortIndexed::Data &out) const {
        const int SIZE = CHUNK_LENGTH + 3;
        std::array<dualmc::DualMC<float>::Point, SIZE * SIZE * SIZE> grid;
        {
            for (int z = 0; z < SIZE; z++) {
            for (int y = 0; y < SIZE; y++) {
            for (int x = 0; x < SIZE; x++) {
                auto &cell = grid[((z * SIZE) + y) * SIZE + x];
                const auto &chunk = surrounding[(z >= CHUNK_LENGTH) + (y >= CHUNK_LENGTH)*2 + (x >= CHUNK_LENGTH)*4];
                const auto &voxel = chunk->get(glm::toIdx(x % CHUNK_LENGTH, y % CHUNK_LENGTH, z % CHUNK_LENGTH));
                cell.x = voxel.density() * 1.f / world::Voxel::DENSITY_MAX;
                cell.w = voxel.material();
            }}}
        }
        {
            std::vector<dualmc::Vertex> dmc_vertices;
            std::vector<dualmc::Tri> dmc_tris;

            dualmc::DualMC<float> builder;
            builder.buildTris(&grid.front(), SIZE, SIZE, SIZE, iso, world::materials::roughness.cbegin(), manifold, dmc_vertices, dmc_tris);

            out.vertices.reserve(dmc_vertices.size());
            out.materials.reserve(dmc_vertices.size());
            out.normals.reserve(dmc_vertices.size());
            for (const auto& v: dmc_vertices) {
                out.vertices.emplace_back(v.x, v.y, v.z);
                out.materials.emplace_back(v.w);
                out.normals.emplace_back(0);
            }

            out.indices.reserve(dmc_tris.size());
            for (const auto& t: dmc_tris) {
                out.indices.push_back(t.i0);
                out.indices.push_back(t.i1);
                out.indices.push_back(t.i2);

                glm::vec3 edge1 = out.vertices[t.i1] - out.vertices[t.i0];
                glm::vec3 edge2 = out.vertices[t.i2] - out.vertices[t.i0];
                glm::vec3 normal = glm::normalize(glm::cross(edge1, edge2));

                out.normals[t.i0] += normal;
                out.normals[t.i1] += normal;
                out.normals[t.i2] += normal;
            }

            for (auto &n : out.normals) {
                n = glm::normalize(n);
            }
        }
    }
}
