#include "Renderer.hpp"

#include "../world/materials.hpp"
#include "../control/Camera.hpp"

Renderer::Renderer(const Renderer::options& options) {
    glGenVertexArrays(1, &VertexArrayID);
    glBindVertexArray(VertexArrayID);

    MainPass = new pass::MainProgram(options.main);
    SkyPass = new pass::SkyProgram();
    SkyEnable = options.skybox;

    FogColor = glm::vec3(options.clear_color.x, options.clear_color.y, options.clear_color.z);
    loadTextures(options.textures, options.mipMapLOD);
}

Renderer::~Renderer() {
    unloadTextures();
    delete MainPass;
    delete SkyPass;
    glDeleteVertexArrays(1, &VertexArrayID);
}

pass::Context Renderer::getPass() {
    glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
    return pass::Context(this, MainPass);
}
void Renderer::postProcess() {
    if(SkyEnable) {
        SkyPass->draw(this);
    }
}

void Renderer::reloadShaders(const pass::MainProgram::options& options) {
    delete MainPass;
    MainPass = new pass::MainProgram(options);
}
void Renderer::reloadTextures(const std::string& texturePath, float mipMapLOD) {
    unloadTextures();
    loadTextures(texturePath, mipMapLOD);
}

void Renderer::unloadTextures() {
    glDeleteTextures(1, &HOSAtlas);
    glDeleteTextures(1, &NormalAtlas);
    glDeleteTextures(1, &TextureAtlas);
}
void Renderer::loadTextures(const std::string& texturePath, float mipMapLOD) {
    std::vector<std::string> terrainTextures;
    for(const auto texture: world::materials::textures) {
        terrainTextures.emplace_back(texturePath + "/terrain/" + texture);
    }
    TextureAtlas = pass::Program::loadTextureArray(terrainTextures, "", mipMapLOD);
    NormalAtlas = pass::Program::loadTextureArray(terrainTextures, ".nrm", mipMapLOD);
    HOSAtlas = pass::Program::loadTextureArray(terrainTextures, ".hos", mipMapLOD);

    Skybox = pass::Program::loadTextureCube(texturePath + "/sky/Space_tray");
}

void Renderer::lookFrom(const Camera& camera) {
    ProjectionMatrix = camera.getProjectionMatrix();
    ViewMatrix = camera.getViewMatrix();
    FogDepth = camera.getDepth();
}