#include "ShortIndexed.hpp"

#include "vboindexer.hpp"
#include "../../data/logger.hpp"

using namespace buffer;

ShortIndexed::Data::Data(const std::vector<VertexData> &vs, const std::vector<GLushort> &indices): indices(indices) {
    vertices.reserve(vs.size());
    materials.reserve(vs.size());
    normals.reserve(vs.size());
    for (auto vertex : vs) {
        vertices.push_back(vertex.Position);
        materials.push_back(vertex.Material);
        normals.push_back(vertex.Normal);
    }
}
void ShortIndexed::Data::index(const std::vector<VertexData>& vs) {
    indexVBO(vs, indices, vertices, materials, normals);
}

ShortIndexed::ShortIndexed(GLenum shape, const std::vector<VertexData> &vertices): Abstract(shape) {
    setData(ShortIndexed::Data(vertices));
}
ShortIndexed::ShortIndexed(GLenum shape, const std::vector<VertexData> &vertices, const std::vector<GLushort>& indices): Abstract(shape) {
    setData(ShortIndexed::Data(vertices, indices));
}
ShortIndexed::ShortIndexed(GLenum shape, const ShortIndexed::Data &data): Abstract(shape) {
    setData(data);
}

ShortIndexed::~ShortIndexed() {
    glDeleteBuffers(1, &NormalBufferID);
    glDeleteBuffers(1, &MaterialBufferID);
    glDeleteBuffers(1, &IndexBufferID);
}

void ShortIndexed::enableAllAttribs() {
    enableVertexAttrib();

    glEnableVertexAttribArray(1);
    glBindBuffer(GL_ARRAY_BUFFER, MaterialBufferID);
    glVertexAttribIPointer(
        1,        // attribute
        1,        // size
        GL_UNSIGNED_SHORT, // type
        0,        // stride
        (void *)0 // array buffer offset
    );

    glEnableVertexAttribArray(2);
    glBindBuffer(GL_ARRAY_BUFFER, NormalBufferID);
    glVertexAttribPointer(
        2,        // attribute
        3,        // size
        GL_FLOAT, // type
        GL_FALSE, // normalized?
        0,        // stride
        (void *)0 // array buffer offset
    );
}
void ShortIndexed::disableAllAttribs() {
    glDisableVertexAttribArray(2);
    glDisableVertexAttribArray(1);
    disableVertexAttrib();
}
void ShortIndexed::enableIndex() {
    glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, IndexBufferID);
}

uint ShortIndexed::draw(buffer::params params) {
    if(IndexSize == 0)
        return 0;

    if(params.vertexOnly) {
        enableVertexAttrib();
    } else {
        enableAllAttribs();
    }
    enableIndex();
    glDrawElements(Shape, IndexSize, GL_UNSIGNED_SHORT, (void *)0);

    if(params.vertexOnly) {
        disableVertexAttrib();
    } else {
        disableAllAttribs();
    }
    return IndexSize;
}

#include <iostream>
void ShortIndexed::setData(const ShortIndexed::Data& data) {
    glGenBuffers(1, &IndexBufferID);
    glGenBuffers(1, &MaterialBufferID);
    glGenBuffers(1, &NormalBufferID);

    IndexSize = data.indices.size();
    if(IndexSize != data.indices.size()) {
        LOG_E("ShortBuffer overflow: " << data.indices.size());
    }
    setIndicies(IndexSize * sizeof(GLushort), &data.indices[0]);
    setVertices(data.vertices.size() * sizeof(glm::vec3), &data.vertices[0]);
    setMaterials(data.materials.size() * sizeof(GLushort), &data.materials[0]);
    setNormals(data.normals.size() * sizeof(glm::vec3), &data.normals[0]);
}

void ShortIndexed::setIndicies(const unsigned long size, const void *data) {
    glBindBuffer(GL_ARRAY_BUFFER, IndexBufferID);
    glBufferData(GL_ARRAY_BUFFER, size, data, GL_STATIC_DRAW);
}
void ShortIndexed::setMaterials(const unsigned long size, const void *data) {
    glBindBuffer(GL_ARRAY_BUFFER, MaterialBufferID);
    glBufferData(GL_ARRAY_BUFFER, size, data, GL_STATIC_DRAW);
}
void ShortIndexed::setNormals(const unsigned long size, const void *data) {
    glBindBuffer(GL_ARRAY_BUFFER, NormalBufferID);
    glBufferData(GL_ARRAY_BUFFER, size, data, GL_STATIC_DRAW);
}